"""
Test the Gaussian Processes in bayesvalidrox

Class GaussianProcessSklearn
    build_meta_model       - x
    check_is_gaussian      - x
    build_kernels          - x
    transform_x            - x
    fit                    - x 
    adaptive_regression    - x
    scale_x                - x 
    eval_meta_model        - x
    calculate_moments      - x
"""

import numpy as np
import pytest
import sys

from sklearn.preprocessing import MinMaxScaler, StandardScaler
from sklearn.gaussian_process import kernels

sys.path.append("../src/")

from bayesvalidrox import GPESkl, Input
from bayesvalidrox.surrogate_models.input_space import InputSpace

@pytest.fixture
def GPE(): 
    
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    samples = np.array([[0.2], [0.8]])
    MetaModel = GPESkl(input)
    MetaModel.CollocationPoints = samples
    MetaModel.InputSpace = InputSpace(MetaModel.input_obj, MetaModel.meta_model_type)
    MetaModel.InputSpace.init_param_space(np.max(MetaModel._pce_deg))
    return MetaModel

def test_meta_model():
    """Construct GPSklearn without input"""
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    GPESkl(input)

def test_build_metamodel(GPE) -> None:
    """
    Build GPE
    """
    MetaModel = GPE
    MetaModel.CollocationPoints = np.array([[0.2], [0.8]])
    MetaModel.build_metamodel()

def test_check_is_gaussian(GPE) -> None:
    """
    Check if Gaussian
    """
    MetaModel = GPE
    MetaModel.check_is_gaussian(1)
    assert MetaModel.is_gaussian == True, "Expected is_gaussian to be True"
    
def test_add_inputspace() -> None:
    """
    Add InputSpace in PCE
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.add_InputSpace()

def test_add_inputspace() -> None:
    """
    Add InputSpace in PCE
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.add_InputSpace()

def test_fit() -> None:
    """
    Fit GPE
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})

# def test_fit_parallel() -> None:
#     """
#     Fit GPE in parallel
#     """
#     inp = Input()
#     inp.add_marginals()
#     inp.Marginals[0].dist_type = 'normal'
#     inp.Marginals[0].parameters = [0, 1]
#     mm = GPESkl(inp)
#     mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, parallel=True)

def test_fit_verbose() -> None:
    """
    Fit PCE verbose
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]}, verbose=True)

def test_fit_pca() -> None:
    """
    Fit PCE verbose and with pca
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(inp)
    MetaModel.dim_red_method = 'pca'
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})

def test_build_kernels(GPE) -> None:
    """
    Build kernels
    """
    MetaModel = GPE
    MetaModel.build_kernels()

def test_kernel_no_kertyp(GPE) -> None:
    """
    Test kernel with no kernel type
    """
    MetaModel = GPE
    MetaModel.build_kernels()
    assert MetaModel._kernel_type == 'RBF'

def test_build_kernel_wrong_kername(GPE) -> None:
    """
    Test kernel with an invalid kernel name
    """
    MetaModel = GPE
    MetaModel._kernel_type = 'InvalidKernel'

    with pytest.raises(AttributeError, match="The kernel option InvalidKernel is not available."):
        MetaModel.build_kernels()

def test_build_kernel_wrong_kertyp(GPE) -> None:
    """
    Test building kernels with an invalid variable type
    """
    MetaModel = GPE
    MetaModel._kernel_type = 123

    with pytest.raises(TypeError, match="The kernel option 123 is of an invalid type."):
        MetaModel.build_kernels()

def test_kernel_type_rbf(GPE) -> None:
    """
    Test kernel type
    """
    MetaModel = GPE
    MetaModel._kernel_type = 'RBF'
    kernel_list, kernel_names = MetaModel.build_kernels()

    assert len(kernel_list) == 1, "Expected only one kernel when autoSelect is False"
    assert kernel_names == ['RBF'], "Expected kernel name to be 'RBF'"

def test_build_kernels_matern(GPE) -> None:
    """
    Build kernels with Matern kernel type
    """
    MetaModel = GPE
    MetaModel._kernel_type = 'Matern'
    kernel_list, kernel_names = MetaModel.build_kernels()

    assert len(kernel_list) == 1, "Expected only one kernel when autoSelect is False"
    assert kernel_names == ['Matern'], "Expected kernel name to be 'Matern'"

def test_build_kernels_rq(GPE) -> None:
    """
    Build kernels with Matern kernel type
    """
    MetaModel = GPE
    MetaModel._kernel_type = 'RQ'
    kernel_list, kernel_names = MetaModel.build_kernels()

    assert len(kernel_list) == 1, "Expected only one kernel when autoSelect is False"
    assert kernel_names == ['RQ'], "Expected kernel name to be 'RQ'"

def test_auto_select_kernels(GPE) -> None:
    """Build Kernels when autoSelect is True"""
    MetaModel = GPE
    MetaModel._autoSelect = True
    kernel_list, kernel_names = MetaModel.build_kernels()

    assert len(kernel_list) == 3, "Expected three kernels when autoSelect is True"
    assert sorted(kernel_names) == ['Matern', 'RBF', 'RQ'], "Expected kernel names to be 'Matern', 'RBF', 'RQ'"

def test_build_kernels_with_length_scale(GPE) -> None:
    """
    Build kernels with specific length scale
    """
    MetaModel = GPE
    MetaModel.kernel_length_scale = 1.5
    MetaModel.build_kernels()

def test_build_kernels_with_bounds(GPE) -> None:
    """
    Build kernels with specific bounds
    """
    MetaModel = GPE
    MetaModel.kernel_bounds = (1e-2, 1e1)
    MetaModel.build_kernels()

def test_kernel_isotropy(GPE):
    """
    Test kernel isotropy
    """
    MetaModel = GPE
    MetaModel.kernel_isotropy = True
    kernels_list, kernel_names = MetaModel.build_kernels()

    assert len(kernels_list) == 1, "Expected only one kernel when autoSelect is False"
    assert isinstance(kernels_list[0].k2.length_scale, int), f"Expected 1 length scales for isotropic kernel, but got a list of them"

def test_anisotropic_kernel(GPE) -> None:
    """Build anisotropic kernels for a 2d case"""
    ndim = 2
    Inputs = Input()
    for i in range(ndim):
        Inputs.add_marginals()
        Inputs.Marginals[i].dist_type = 'uniform'
        Inputs.Marginals[i].parameters = [0, 1]
    samples = np.array([[0.2, 0.5], [0.8, 0.7]])
    MetaModel = GPESkl(Inputs)
    MetaModel.CollocationPoints = samples
    MetaModel.InputSpace = InputSpace(MetaModel.input_obj, MetaModel.meta_model_type)
    MetaModel._kernel_isotropy = False
    MetaModel._kernel_type = 'RBF'

    kernels_list, kernel_names = MetaModel.build_kernels()

    assert len(kernels_list) == 1, "Expected only one kernel when autoSelect is False"
    assert kernel_names == ['RBF'], "Expected kernel name to be 'RBF'"
    assert len(kernels_list[
                   0].k2.length_scale) == ndim, f"Expected {ndim} length scales for anisotropic kernel, but got {len(kernels_list[0].k2.length_scale)}"

def test_adaptive_regression(GPE) -> None: 
    """
    Regression without a specific method 
    """
    MetaModel = GPE
    samples = np.array([[0.0], [0.1]])
    outputs = np.array([[0.1], [0.1]])

    MetaModel.build_metamodel()

    MetaModel.adaptive_regression(samples, outputs, varIdx=None)

def test_adaptive_regression_verbose(GPE) -> None: 
    """
    Regression without a specific method with verbose
    """
    MetaModel = GPE
    samples = np.array([[0.0], [0.1]])
    outputs = np.array([[0.1], [0.1]])

    MetaModel.build_metamodel()

    MetaModel.adaptive_regression(samples, outputs, varIdx=None, verbose=True)


# Added:
def test_transform_scale_x_norm(GPE) -> None:
    """
    Test normalization using 'norm' (MinMaxScaler), for both the transform_x() and scale_x() functions.
    """
    MetaModel = GPE
    X = np.array([[1, 2], [3, 4], [5, 6]])
    X_transformed, scaler = MetaModel.transform_x(X, transform_type='norm')  # Call directly on class

    expected_scaler = MinMaxScaler()
    expected_X_transformed = expected_scaler.fit_transform(X)

    assert np.allclose(X_transformed, expected_X_transformed), "Normalization failed"
    assert isinstance(scaler, MinMaxScaler), "Scaler object is not MinMaxScaler"

    MetaModel._x_scaler = scaler
    X_scaled = MetaModel.scale_x(X, MetaModel._x_scaler)
    assert np.allclose(X_scaled, expected_X_transformed), "Scaling after normalization failed"


def test_transform_scale_x_standard(GPE) -> None:
    """Test standardization using 'standard' from the Scikit-Learn library, for both the transformation and the
    scaling functions"""
    X = np.array([[1, 2], [3, 4], [5, 6]])
    MetaModel = GPE
    X_transformed, scaler = MetaModel.transform_x(X, transform_type='standard')

    expected_scaler = StandardScaler()
    expected_X_transformed = expected_scaler.fit_transform(X)

    assert np.allclose(X_transformed, expected_X_transformed), "Standardization failed"
    assert isinstance(scaler, StandardScaler), "Scaler object is not StandardScaler"

    MetaModel._x_scaler = scaler
    X_scaled = MetaModel.scale_x(X, MetaModel._x_scaler)
    assert np.allclose(X_scaled, expected_X_transformed), "Scaling after standardization failed"


def test_transform_x_none(GPE) -> None:
    """Test the no transformation case when transform_type is None"""
    X = np.array([[1, 2], [3, 4], [5, 6]])
    MetaModel = GPE
    X_transformed, scaler = MetaModel.transform_x(X, transform_type=None)

    assert np.array_equal(X_transformed, X), "No transformation failed"
    assert scaler is None, "Scaler object should be None"

    # Test scale_x(scaler=None)
    MetaModel._x_scaler = scaler
    X_scaled = MetaModel.scale_x(X, MetaModel._x_scaler)
    assert np.allclose(X_scaled, X), "Scaling after standardization failed"


def test_transform_x_invalid_type(GPE) -> None:
    """Test transformation with an invalid transformation type"""
    X = np.array([[1, 2], [3, 4], [5, 6]])
    MetaModel = GPE

    # Expecting an AttributeError when an invalid transformation type is provided
    with pytest.raises(AttributeError, match="No scaler invalid found."):
        MetaModel.transform_x(X, transform_type='invalid')

    stop = 1


def test_scale_x_none(GPE) -> None:
    X = np.array([[1, 2], [3, 4], [5, 6]])
    MetaModel = GPE
    _, scaler = MetaModel.transform_x(X, transform_type=None)
    MetaModel._x_scaler = scaler

    X_transformed = MetaModel.scale_x(X, scaler)
    assert np.array_equal(X_transformed, X), "No transformation failed"

#TO Do: 
def test_scale_x(GPE):
    """
    Scale x
    """
    MetaModel = GPE
    x_s, scaler = MetaModel.transform_x(np.array([[0.1], [0.2]]))
    MetaModel.scale_x(np.array([[0.1], [0.2]]), scaler)


def test_eval_metamodel() -> None:
    """
    Evaluate trained GPE
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.out_names = ['Z']
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    MetaModel.eval_metamodel([[0.4]])


def test_eval_metamodel_multiple_outputs() -> None:
    """
    Evaluate trained GPE with multiple outputs
    """
    input = Input()
    input.add_marginals()
    input.Marginals[0].dist_type = 'normal'
    input.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(input)
    MetaModel.out_names = ['Z1', 'Z2']
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z1': [[0.4], [0.2], [0.5]], 'Z2': [[0.3], [0.1], [0.4]]})
    MetaModel.eval_metamodel([[0.4]])

def test_eval_metamodel_normalboots() -> None:
    """
    Eval trained PCE with normal bootstrap
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.bootstrap_method = 'normal'
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.eval_metamodel([[0.4]])


def test_eval_metamodel_highnormalboots() -> None:
    """
    Eval trained PCE with higher bootstrap-itrs
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.n_bootstrap_itrs = 2
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    mm.eval_metamodel([[0.4]])


def test_eval_metamodel_pca() -> None:
    """
    Eval trained PCE with pca
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    mm = GPESkl(inp)
    mm.dim_red_method = 'pca'
    mm.out_names = ['Z']
    mm.fit([[0.2], [0.8]], {'Z': [[0.4, 0.4], [0.5, 0.6]]})
    mm.eval_metamodel([[0.4]])


def test_pca_transformation() -> None:
    """
    Apply PCA
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(inp)
    outputs = np.array([[0.4, 0.4], [0.5, 0.6]])
    MetaModel.pca_transformation(outputs, 1)



def test_pca_transformation_varcomp() -> None:
    """
    Apply PCA with set var_pca_threshold
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(inp)
    outputs = np.array([[0.4, 0.4], [0.5, 0.6]])
    MetaModel.var_pca_threshold = 1
    MetaModel.pca_transformation(outputs, 1)


#%% Test GPE.auto_vivification
def test_auto_vivification() -> None:
    """
    Creation of auto-vivification objects
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(inp)
    MetaModel.auto_vivification()


#%% Test GPE.copy_meta_model_opts

def test_copy_meta_model_opts() -> None:
    """
    Copy the PCE with just some stats
    """
    inp = Input()
    inp.add_marginals()
    inp.Marginals[0].dist_type = 'normal'
    inp.Marginals[0].parameters = [0, 1]
    MetaModel = GPESkl(inp)
    MetaModel.add_InputSpace()
    MetaModel.copy_meta_model_opts()

def test_eval_metamodel_with_pca(GPE) -> None:
    """
    Evaluate trained GPE with PCA
    """
    MetaModel = GPE
    MetaModel.out_names = ['Z']
    MetaModel.dim_red_method = 'pca'
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    MetaModel.eval_metamodel([[0.4]])

def test_eval_metamodel_invalid_input(GPE) -> None:
    """
    Evaluate trained GPE with invalid input
    """
    MetaModel = GPE
    MetaModel.out_names = ['Z']
    MetaModel.fit([[0.2], [0.4], [0.8]], {'Z': [[0.4], [0.2], [0.5]]})
    with pytest.raises(AttributeError):
        MetaModel.eval_metamodel([[0.4, 0.5]])
